/*
 * crashallassrt: Walks all the tasks' stacks, both kernel and user,
 * inorder to capture the back traces
 *
 * Copyright (C) 2005, 2006 TiVo Inc. All Rights Reserved.
 */

#include <linux/string.h>
#include <asm/uaccess.h>
#include <asm/param.h>
#include <asm/user.h>
#include <linux/sched.h>
#include <linux/threads.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/tivo-crashallassert.h>
#include <linux/fastbufs.h>

extern int rt_scheduler_disabled;
extern void show_trace_task_bt(struct task_struct *tsk, int dispconsole);
extern void do_emergency_sync(void);
extern void tivo_restart(char *command);

extern struct module kernel_module;

void stop_runnables(void);
void restore_runnables(void);
static ssize_t pid_read_maps (struct task_struct *, char *);

/*
 * The max number is PID_MAX defined in linux/threads.h
 */
#define THREAD_MAP_BLKS         (PID_MAX/(sizeof(u_long)*8))
u_int *tgl_bit_map=NULL;
u_int *runq_bit_map=NULL;

static int debugconsole = 1;
#define DEBUG_LEVEL  KERN_ERR

static ssize_t pid_read_maps (struct task_struct *task, char *tmp)
{
	struct mm_struct *mm;
	struct vm_area_struct * map;
	long retval;

	task_lock(task);
	mm = task->mm;
	if (mm)
		atomic_inc(&mm->mm_users);
	task_unlock(task);

	retval = 0;
	if (!mm)
		return retval;

	down_read(&mm->mmap_sem);
	map = mm->mmap;

	while (map) {
        if ( (map->vm_file == NULL) || 
                !((map->vm_flags & VM_READ) && 
                (map->vm_flags & VM_EXEC))) { 

        map = map->vm_next;
                continue;
        }
        char *line = d_path(map->vm_file->f_dentry,
                            map->vm_file->f_vfsmnt,
                            tmp, PAGE_SIZE);
        printk( DEBUG_LEVEL "read 0x%08lx 0x%08lx %s\n",
                            map->vm_start,map->vm_end,line);
		map = map->vm_next;
	}
	up_read(&mm->mmap_sem);
	mmput(mm);

	return retval;
}

void restore_runnables(){
    u_int *tmp;
    struct task_struct *tsk; 
    struct task_struct *t; 
    int shift = 0;
    int seg;

    for_each_task(tsk) {
        if (    tsk->mm && 
                tsk != &init_task &&
                tsk != current &&
                strcmp(tsk->comm,"syslogd") && 
                strcmp(tsk->comm,"klogd") && 
                strcmp(tsk->comm,"init")) {

                for ( tmp = &runq_bit_map[0], seg=0; 
                    seg < THREAD_MAP_BLKS; tmp++,seg++){
                    shift = 0;
                    while( shift < sizeof(u_long)*8 ) {
                        if ( *tmp & (1<<shift) ) {
                            t = find_task_by_pid(
                                shift+
                                seg*sizeof(u_long)*8
                                    );
                            if (t)
                                t->cpus_runnable = 1UL;
                        }
                        shift++;
                    }
                }
        }
    }
}

void stop_runnables(){ 
    u_int *tmp;
    struct task_struct *tsk; 
    for_each_task(tsk) {
        /*
         * donot stop klogd, syslogd, init, current process, 
         * swapper and kernel threads
        */ 
        if (tsk->mm && 
                tsk != &init_task &&
                tsk != current &&
                strcmp(tsk->comm,"syslogd") &&
                strcmp(tsk->comm,"klogd") && 
                strcmp(tsk->comm,"init")) {
                    tmp = &runq_bit_map[tsk->pid/(sizeof(u_long)*8)];
                    *tmp  |= 1<<(tsk->pid%(sizeof(u_long)*8));
                    tsk->cpus_runnable = 0;
                }
    }
}

int tivocrashallassert(unsigned long arg){
    struct task_struct *tsk; 
    u_int *tmp;
    char *tbuf;
    int seg;
    int shift;
    struct k_arg_list k_arg_list;
    struct k_arg *k_args,*k_argsp;
    unsigned int time_stamp = 0xdeadbeaf;
    int err;
    int count;
    int tmkdbg=0;
    int retval = 0;
    struct module *mod;
    char mod_path[256];
    int i;

    err = copy_from_user((void *)&k_arg_list,
                    (void *)arg,sizeof(struct k_arg_list));
    if ( err )
        return -EFAULT;

    k_args = (struct k_arg *)
                kmalloc(sizeof(struct k_arg)*k_arg_list.count,GFP_KERNEL);
    if ( k_args == NULL )
        return -ENOMEM;

    err = copy_from_user((void *)k_args,
                            (void *)k_arg_list.kargs,
                            sizeof(struct k_arg)*k_arg_list.count);
    if (err){
        kfree(k_args);
        return -EFAULT;
    }

    for ( count=0,k_argsp=k_args; count < k_arg_list.count; k_argsp++,count++){
       switch(k_argsp->tag){
            case TAG_TIME:
                time_stamp = k_argsp->ptr;
            break;
            case TAG_TMKDBG:
                if ( k_argsp->ptr ) 
                    tmkdbg = 1;
                else
                    tmkdbg = 0; 
            break; 
            default:
                printk( DEBUG_LEVEL "Invalid tag/ptr/len\n");
       }
    }

    kfree(k_args);

	tbuf = (char*)__get_free_page(GFP_KERNEL);
	if (!tbuf)
        return -ENOMEM;

    tgl_bit_map =(u_int *)kmalloc(THREAD_MAP_BLKS,GFP_KERNEL);
    if ( tgl_bit_map == NULL ){
        retval = -ENOMEM;
        goto out;
    }

    runq_bit_map =(u_int *)kmalloc(THREAD_MAP_BLKS,GFP_KERNEL);
    if ( runq_bit_map == NULL ){
        retval = -ENOMEM;
        goto out;
    }
        
    /*
     *Dsable the Real Time scheduler
     */ 
    rt_scheduler_disabled = 1;

    /*
     *Prevent allmost all runnable threads not to get scheduled on cpu
     */ 
    stop_runnables();

    printk ( DEBUG_LEVEL "bt_all_thread 0x%08x start\n",time_stamp);

    printk( DEBUG_LEVEL "global\n");

    /*Kernel Loadable Module information*/
	for (mod = module_list; mod ; mod = mod->next) {
		if (mod->flags & MOD_DELETED || mod == &kernel_module)
			continue;
        for ( i = mod->nsyms -1; i >= 0; i--){
            if ( strstr(mod->syms[i].name,"/lib/modules/")){
                if ( strstr(mod->syms[i].name,"/platform/lib/modules/"))
                    strcpy(mod_path,"/platform/lib/modules/");
                else
                    strcpy(mod_path,"/lib/modules/");

                strcat(mod_path,mod->name);
                strcat(mod_path,".o");
                printk(DEBUG_LEVEL "read 0x%08x %s\n",(unsigned long) mod, mod_path);
                break;
            }
        }
	}

    /*
     *Find thread group leaders
     */ 
    for_each_task(tsk) {
        tmp = &tgl_bit_map[tsk->pgrp/(sizeof(u_long)*8)];
            *tmp  |= 1<<(tsk->pgrp%(sizeof(u_long)*8));
    }

    for ( tmp = &tgl_bit_map[0], seg=0; seg < THREAD_MAP_BLKS; tmp++,seg++){
        shift = 0;
        while( shift < sizeof(u_long)*8 ) {
            if ( *tmp & (1<<shift) ) {
                int pgrp = shift+seg*sizeof(u_long)*8;
                int leader = 1;    
                for_each_task(tsk){
                    if ( pgrp == tsk->pgrp){
                        if ( leader){
                            printk(DEBUG_LEVEL "threadgroup %d name \"%s\"\n",
                                                        tsk->pgrp,tsk->comm);

                            /*
                             *Dump memory map of a valid threadgroup
                            */ 
                            pid_read_maps(tsk, tbuf);
                            leader = 0;
                        }
                                                
                        printk(DEBUG_LEVEL "pid %d threadgroup %d name \"%s",
                            tsk->pid, 
                            tsk->pgrp,
                            tsk->comm);

                        thread_fbuf *exec_cntxt = (thread_fbuf *) tsk->exec_cntxt;
                        if ( exec_cntxt && exec_cntxt->cntxt )
                            printk(DEBUG_LEVEL "-%s\"\n",exec_cntxt->cntxt->name);
                        else
                            printk(DEBUG_LEVEL "\"\n");
                                
                        /*
                         *Dump back traces 
                         */ 

                        show_trace_task_bt(tsk,debugconsole);
                                            
                        /*let other threads like syslogd run*/
                        schedule();
                    }
                }
            }
            shift++;
        }
    }

    printk ( DEBUG_LEVEL "bt_all_thread 0x%08x end\n",time_stamp);

	free_page((unsigned long)tbuf);
        
    /*
     * Let runnable threads get scheduled as usual but with 
     * rt scheduler still turned off
     */ 
    restore_runnables();

    if ( !tmkdbg ){
        /*
         * flush all disk buffers
         */ 
        do_emergency_sync();

        /*
         *Done dumping back traces. Reboot the TCD
         */ 
        tivo_restart(NULL);
    }
out:
    if (tgl_bit_map)
        kfree(tgl_bit_map);
    if (runq_bit_map)
        kfree(runq_bit_map);
    return retval;
}

#ifdef CONFIG_TICK_KILLER
int tick_killer_start(int ticks) {
	if(ticks >= 0) {
		if(tick_killer_ticks == ticks) {
			printk("TickKiller called with %d, no effect\n", ticks);
			return 0;
		}
		if(ticks == TICK_KILLER_DISABLED) {
			printk("TickKiller disabled\n");
		} else {
			printk("TickKiller will kill the TCD in %d ticks\n",
					ticks);
		}
		tick_killer_ticks = ticks;
	} else {
		return -EINVAL;
	}
	return 0;
}
#endif
